import dataclasses
import os
import shutil
import traceback
from copy import deepcopy
from datetime import datetime
from typing import List, Optional

import yaml
from vipen.core.logging import console
from vipen.core.task import BaseError
from vipen.core.utils import dataclass
from vipen.monitoring.monitoring import Monitoring
from vipen.providers.base import Domain, Provider, Server
from vipen.templates.base import Template

DEFAULT_CONFIG_DIR = os.path.expanduser("~/.vipen")


@dataclass
class Cluster:
    name: str
    servers: List[Server]
    domains: List[Domain]
    template: Template

    def get_server(self, name=None, id=None) -> Server:
        if name is None and id is None:
            raise ValueError("Must specify either name or id")

        for s in self.servers:
            if (name is None or s.name == name) and (id is None or s.id == id):
                return s
        return None

    def get_domain(self, name) -> Domain:
        for d in self.domains:
            if d.name == name:
                return d
        return None

    @classmethod
    def new(cls, name):
        return cls(
            name=name,
            servers=[],
            domains=[],
            template=None,
        )


@dataclass
class Config:
    providers: List[Provider]
    clusters: List[Cluster]
    current_cluster: Cluster
    monitoring: Monitoring

    provider_cls = []
    template_cls = []

    def get_provider(self, name) -> Provider:
        for p in self.providers:
            if p.name == name:
                return p
        return None

    def get_cluster(self, name) -> Cluster:
        for c in self.clusters:
            if c.name == name:
                return c
        return None

    def get_all_servers(self) -> List[Server]:
        servers = []
        for c in self.clusters:
            servers.extend(c.servers)
        return servers

    def get_all_domains(self) -> List[Domain]:
        domains = []
        for c in self.clusters:
            domains.extend(c.domains)
        return domains

    @property
    def server_providers(self):
        return [p for p in self.providers if p.can_server]

    @property
    def cdn_providers(self):
        return [p for p in self.providers if p.can_cdn]

    def asdict(self):
        return dataclasses.asdict(self, dict_factory=self.dict_factory)

    def asjson(self, *args, **kwargs):
        import json

        return json.dumps(self.asdict(), *args, **kwargs)

    def asyaml(self, *args, **kwargs):
        import yaml

        return yaml.safe_dump(self.asdict(), *args, **kwargs)

    def dict_factory(self, args):
        def clean(k, v):
            if k.startswith("_"):
                return None
            if k == "provider":
                return k, v["name"]
            if k == "current_cluster":
                return "default_cluster", v["name"] if v else ""
            return k, v

        args = [clean(*a) for a in args]
        return dict([a for a in args if a is not None])

    def save(self, path):
        # Backup existing config incase we mess things up when saving new one
        backup_dir = os.path.join(DEFAULT_CONFIG_DIR, "backups")
        if not os.path.exists(backup_dir):
            os.makedirs(backup_dir)
        if os.path.exists(path):
            bkup = os.path.join(
                backup_dir, f"config-{datetime.now().strftime('%Y%m%d-%H%M%S')}.yaml"
            )
            shutil.copyfile(path, bkup)

        try:
            data = self.asyaml(sort_keys=False)
        except Exception as e:
            console.error(traceback.format_exc())
            console.error("Saving config failed")
            return

        with open(path, "w") as f:
            f.write(data)


class ConfigError(BaseError):
    pass


def load_config(path, use_cluster=None, provider_cls=[], template_cls=[]):
    config = Config(providers=[], clusters=[], current_cluster=None, monitoring=None)
    config.provider_cls = provider_cls
    config.template_cls = template_cls

    if os.path.exists(path):
        with open(path) as f:
            data = yaml.safe_load(f.read())
    else:
        console.log(f"Starting fresh config at {path}")
        data = {}

    original_data = deepcopy(data)

    config.providers = parse_providers(provider_cls, data.get("providers", []))
    config.clusters = parse_clusters(config, template_cls, data.get("clusters", []))

    if use_cluster is None:
        use_cluster = data.get("default_cluster") or None
    if use_cluster is not None and config.get_cluster(use_cluster) is None:
        raise ConfigError(f"Cluster '{use_cluster}' does not exist")
    config.current_cluster = config.get_cluster(use_cluster)

    config.monitoring = parse_monitoring(config, data.get("monitoring"))

    # if config.asdict() != original_data:
    #     diff = DeepDiff(original_data, config.asdict())
    #     print(config.asdict())
    #     print(diff.pretty())
    #     raise ConfigError("Config dump mismatch")

    return config


def parse_providers(provider_cls, data):
    providers = []
    for p in data:
        for cls in provider_cls:
            if p.get("cloud") == cls.cloud:
                providers.append(cls(**p))
                break
        else:
            console.warn(
                f"Provider '{p.get('name')}' has an unknown cloud '{p.get('cloud')}', skipping provider."
            )
    return providers


def parse_clusters(config: Config, template_cls, data):
    clusters = []
    for c in data:
        try:
            clusters.append(parse_cluster(config, template_cls, c))
        except ConfigError as e:
            console.warn(
                f"Unable to parse cluster '{c.get('name')}': {str(e)}. Skipping cluster"
            )
    return clusters


def parse_cluster(config: Config, template_cls, data):
    if not data.get("name"):
        raise ConfigError("Missing name")

    cluster = Cluster.new(data.get("name"))

    for s in data.get("servers", []):
        try:
            cluster.servers.append(parse_server(config, s))
        except ConfigError as e:
            console.warn(
                f"Unable to parse server '{s.get('name')}': {str(e)}. Skipping server"
            )

    for d in data.get("domains", []):
        try:
            cluster.domains.append(parse_domain(config, d))
        except ConfigError as e:
            console.warn(
                f"Unable to parse domain '{d.get('name')}': {str(e)}. Skipping domain"
            )

    cluster.template = parse_template(template_cls, data.get("template", {}))

    return cluster


def parse_server(config: Config, data):
    name = data.get("name")
    if not name:
        raise ConfigError("Missing name")

    if not data.get("ip"):
        raise ConfigError("Missing ip")

    provider = config.get_provider(data.get("provider"))
    if provider is None:
        raise ConfigError(f"Unknown provider '{data.get('provider')}'")

    if not provider.can_server:
        raise ConfigError(
            f"Referring to provider '{data.get('provider')}' which does not support servers"
        )

    data["provider"] = provider
    return provider.server_cls(**data)


def parse_domain(config: Config, data):
    name = data.get("name")
    if not name:
        raise ConfigError("Missing name")

    provider = config.get_provider(data.get("provider"))
    if provider is None:
        raise ConfigError(f"Unknown provider '{data.get('provider')}'")

    if not provider.can_cdn:
        raise ConfigError(
            f"Referring to provider '{data.get('provider')}' which does not support domains"
        )

    data["provider"] = provider
    return provider.domain_cls(**data)


def parse_template(template_cls, data) -> Optional[Template]:
    if not data:
        raise ConfigError("Missing template")

    for cls in template_cls:
        if data.get("type") == cls.type:
            return cls(**data)
    else:
        raise ConfigError(f"Template has an unknown type '{data.get('type')}'")
    return None


def parse_monitoring(config: Config, data: dict) -> Optional[Monitoring]:
    if not data:
        return None

    server_data = data.get("server", None)
    server = None
    if server_data:
        try:
            server = parse_server(config, server_data)
        except ConfigError as e:
            console.warn(f"Unable to parse monitoring server config: {str(e)}")
            return None

    domain_data = data.get("domain", None)
    domain = None
    if domain_data:
        try:
            domain = parse_domain(config, domain_data)
        except ConfigError as e:
            console.warn(f"Unable to parse monitoring data config: {str(e)}")
            return None

    return Monitoring(server=server, domain=domain)



import os
import time
from email.generator import Generator
from typing import List

from vipen.core.config import Cluster
from vipen.core.logging import console
from vipen.core.task import with_prefix
from vipen.core.utils import (
    PromptField,
    dataclass,
    ssh_key_fingerprint,
    ssh_key_public_key,
)

import digitalocean

from .base import CDNProvider, ProviderError, Server, ServerProvider


@dataclass
class DigitalOceanServer(Server):
    def destroy(self) -> Generator:
        yield ("Destroying droplet")
        droplet = self.provider.api.get_droplet(self.id)
        droplet.destroy()


@dataclass
class DigitalOceanProvider(ServerProvider, CDNProvider):
    cloud = "DigitalOcean"
    server_cls = DigitalOceanServer

    api_key: str = PromptField()
    project: str = PromptField(required=False)

    _manager = None

    def init_provider(self):
        if not self.ssh_key:
            raise ProviderError("SSH key not specified")
        if not os.path.exists(self.ssh_key_path):
            raise ProviderError(f"SSH Key '{self.ssh_key}' doesn't exist")

        yield "Checking SSH key exists on DigitalOcean"
        key = self.fetch_ssh_key()

        if key is None:
            console.warn(f"SSH key '{self.ssh_key}' not found in DigitalOcean account")
            yield f"Adding SSH key '{self.ssh_key}' to DigitalOcean"
            key = digitalocean.SSHKey(
                token=self.api.token,
                name="vipen",
                public_key=ssh_key_public_key(self.ssh_key_path),
            )
            key.create()
            console.log(f"Created new SSH key {key.id}")

    def create_server(self, cluster: Cluster, name: str):
        yield f"Creating droplet '{name}'"
        droplet = digitalocean.Droplet(
            token=self.api.token,
            name=name,
            region="fra1",
            image="ubuntu-20-04-x64",
            size_slug="s-1vcpu-512mb-10gb",
            ssh_keys=[self.fetch_ssh_key()],
            backups=False,
            tags=[self.cluster_tag(cluster)],
        )
        droplet.create()

        yield f"Waiting for droplet to be ready"
        timeout = 60
        for i in range(timeout):
            time.sleep(1)
            droplet.load()
            if droplet.status == "active":
                break
            yield f"Waiting for droplet to be ready [{i+1}/{timeout}s]"
        else:
            droplet.destroy()
            raise ProviderError("Droplet creation timed out")

        server = self.droplet_to_server(droplet)

        # Wait for the server to finish updating and installing stuff on first boot
        # https://www.digitalocean.com/community/questions/solved-how-to-wait-for-droplet-agent-to-be-installed-to-avoid-apt-lock-issues
        yield from with_prefix("Connecting to server", server.connect())
        yield from with_prefix(
            "Waiting for server to chill",
            server.run(
                """
                for ((i=1;i<=30;i++)); do
                    echo "[$i/30s]"
                    if ls /etc/systemd/system/droplet-agent.services &> /dev/null; then
                        break
                    fi
                    sleep 1
                done
                """
            ),
            style=None,
        )

        return server

    @property
    def api(self):
        if self._manager is None:
            self._manager = digitalocean.Manager(token=self.api_key)
        return self._manager

    @property
    def ssh_key_path(self):
        return os.path.expanduser(self.ssh_key)

    def fetch_cluster_servers(self, cluster: Cluster) -> List[DigitalOceanServer]:
        return [
            self.droplet_to_server(d)
            for d in self.api.get_all_droplets(tag_name=self.cluster_tag(cluster))
        ]

    def fetch_server(self, id: str) -> Server:
        try:
            return self.droplet_to_server(self.api.get_droplet(id))
        except digitalocean.NotFoundError:
            return None

    def fetch_ssh_key(self) -> digitalocean.SSHKey:
        fingerprint = ssh_key_fingerprint(self.ssh_key_path)
        for k in self.api.get_all_sshkeys():
            if k.fingerprint == fingerprint:
                return k
        return None

    def cluster_tag(self, cluster: Cluster):
        return f"vipen-{cluster.name}"

    def droplet_to_server(self, droplet) -> DigitalOceanServer:
        return self.server_cls(
            provider=self, name=droplet.name, ip=droplet.ip_address, id=droplet.id
        )



import json

dashboard = json.dumps(
    {
        "annotations": {
            "list": [
                {
                    "builtIn": 1,
                    "datasource": "-- Grafana --",
                    "enable": True,
                    "hide": True,
                    "iconColor": "rgba(0, 211, 255, 1)",
                    "name": "Annotations & Alerts",
                    "target": {
                        "limit": 100,
                        "matchAny": False,
                        "tags": [],
                        "type": "dashboard",
                    },
                    "type": "dashboard",
                }
            ]
        },
        "description": "VIPI proxy traffic by AS",
        "editable": True,
        "fiscalYearStartMonth": 0,
        "gnetId": 11206,
        "graphTooltip": 0,
        "id": 1,
        "iteration": 1668446221155,
        "links": [],
        "liveNow": False,
        "panels": [
            {
                "aliasColors": {},
                "bars": False,
                "dashLength": 10,
                "dashes": False,
                "datasource": "Prometheus",
                "fieldConfig": {"defaults": {"links": []}, "overrides": []},
                "fill": 1,
                "fillGradient": 0,
                "gridPos": {"h": 6, "w": 24, "x": 0, "y": 0},
                "hiddenSeries": False,
                "id": 2,
                "legend": {
                    "alignAsTable": True,
                    "avg": True,
                    "current": False,
                    "hideEmpty": False,
                    "hideZero": True,
                    "max": False,
                    "min": False,
                    "rightSide": True,
                    "show": False,
                    "sideWidth": 250,
                    "sort": "avg",
                    "sortDesc": True,
                    "total": False,
                    "values": True,
                },
                "lines": True,
                "linewidth": 1,
                "NonePointMode": "None as zero",
                "options": {"alertThreshold": True},
                "percentage": False,
                "pluginVersion": "8.2.6",
                "pointradius": 2,
                "points": False,
                "renderer": "flot",
                "seriesOverrides": [],
                "spaceLength": 10,
                "stack": False,
                "steppedLine": False,
                "targets": [
                    {
                        "expr": 'topk(5, sum(rate(proxy_downstream_total{job="vipen-server",src_as=~"$as",instance=~"$server"}[1m]) * 8) by (src_as, src_as_name))',
                        "intervalFactor": 1,
                        "legendFormat": "{{src_as_name}} (AS{{src_as}})",
                        "refId": "B",
                    }
                ],
                "thresholds": [],
                "timeFrom": None,
                "timeRegions": [],
                "timeShift": None,
                "title": "Downstream Traffic by AS",
                "tooltip": {"shared": False, "sort": 1, "value_type": "individual"},
                "type": "graph",
                "xaxis": {
                    "buckets": None,
                    "mode": "time",
                    "name": None,
                    "show": True,
                    "values": [],
                },
                "yaxes": [
                    {
                        "format": "bps",
                        "label": "",
                        "logBase": 1,
                        "max": None,
                        "min": None,
                        "show": True,
                    },
                    {
                        "format": "short",
                        "label": None,
                        "logBase": 1,
                        "max": None,
                        "min": None,
                        "show": False,
                    },
                ],
                "yaxis": {"align": False, "alignLevel": None},
            },
            {
                "aliasColors": {},
                "bars": False,
                "dashLength": 10,
                "dashes": False,
                "datasource": "Prometheus",
                "fieldConfig": {"defaults": {"links": []}, "overrides": []},
                "fill": 1,
                "fillGradient": 0,
                "gridPos": {"h": 6, "w": 24, "x": 0, "y": 6},
                "hiddenSeries": False,
                "id": 7,
                "legend": {
                    "alignAsTable": True,
                    "avg": True,
                    "current": False,
                    "hideEmpty": False,
                    "hideZero": True,
                    "max": False,
                    "min": False,
                    "rightSide": True,
                    "show": False,
                    "sideWidth": 250,
                    "sort": "avg",
                    "sortDesc": True,
                    "total": False,
                    "values": True,
                },
                "lines": True,
                "linewidth": 1,
                "NonePointMode": "None as zero",
                "options": {"alertThreshold": True},
                "percentage": False,
                "pluginVersion": "8.2.6",
                "pointradius": 2,
                "points": False,
                "renderer": "flot",
                "seriesOverrides": [{"alias": "/.*TX.*/", "transform": "negative-Y"}],
                "spaceLength": 10,
                "stack": False,
                "steppedLine": False,
                "targets": [
                    {
                        "expr": 'topk(5, sum(rate(proxy_upstream_total{job="vipen-server",src_as=~"$as",instance=~"$server"}[1m]) * 8) by (src_as, src_as_name))',
                        "instant": False,
                        "interval": "",
                        "intervalFactor": 1,
                        "legendFormat": "{{src_as_name}} (AS{{src_as}})",
                        "refId": "A",
                    }
                ],
                "thresholds": [],
                "timeFrom": None,
                "timeRegions": [],
                "timeShift": None,
                "title": "Upstream Traffic by AS",
                "tooltip": {"shared": False, "sort": 1, "value_type": "individual"},
                "type": "graph",
                "xaxis": {
                    "buckets": None,
                    "mode": "time",
                    "name": None,
                    "show": True,
                    "values": [],
                },
                "yaxes": [
                    {
                        "format": "bps",
                        "label": "",
                        "logBase": 1,
                        "max": None,
                        "min": None,
                        "show": True,
                    },
                    {
                        "format": "short",
                        "label": None,
                        "logBase": 1,
                        "max": None,
                        "min": None,
                        "show": False,
                    },
                ],
                "yaxis": {"align": False, "alignLevel": None},
            },
            {
                "cacheTimeout": None,
                "columns": [],
                "datasource": "Prometheus",
                "fontSize": "100%",
                "gridPos": {"h": 11, "w": 12, "x": 0, "y": 12},
                "id": 5,
                "links": [],
                "pageSize": None,
                "pluginVersion": "6.3.6",
                "scroll": True,
                "showHeader": True,
                "sort": {"col": 8, "desc": True},
                "styles": [
                    {
                        "alias": "",
                        "align": "auto",
                        "colorMode": None,
                        "colors": [
                            "rgba(245, 54, 54, 0.9)",
                            "rgba(237, 129, 40, 0.89)",
                            "rgba(50, 172, 45, 0.97)",
                        ],
                        "dateFormat": "YYYY-MM-DD HH:mm:ss",
                        "decimals": 2,
                        "mappingType": 1,
                        "pattern": "Time",
                        "thresholds": [],
                        "type": "hidden",
                        "unit": "short",
                    },
                    {
                        "alias": "Bandwidth",
                        "align": "auto",
                        "colorMode": None,
                        "colors": [
                            "rgba(245, 54, 54, 0.9)",
                            "rgba(237, 129, 40, 0.89)",
                            "rgba(50, 172, 45, 0.97)",
                        ],
                        "dateFormat": "YYYY-MM-DD HH:mm:ss",
                        "decimals": 2,
                        "link": False,
                        "linkUrl": "",
                        "mappingType": 1,
                        "pattern": "Value",
                        "thresholds": [],
                        "type": "number",
                        "unit": "bps",
                    },
                    {
                        "alias": "",
                        "align": "auto",
                        "colorMode": None,
                        "colors": [
                            "rgba(245, 54, 54, 0.9)",
                            "rgba(237, 129, 40, 0.89)",
                            "rgba(50, 172, 45, 0.97)",
                        ],
                        "dateFormat": "YYYY-MM-DD HH:mm:ss",
                        "decimals": 2,
                        "mappingType": 1,
                        "pattern": "__name__",
                        "thresholds": [],
                        "type": "hidden",
                        "unit": "short",
                    },
                    {
                        "alias": "",
                        "align": "auto",
                        "colorMode": None,
                        "colors": [
                            "rgba(245, 54, 54, 0.9)",
                            "rgba(237, 129, 40, 0.89)",
                            "rgba(50, 172, 45, 0.97)",
                        ],
                        "dateFormat": "YYYY-MM-DD HH:mm:ss",
                        "decimals": 2,
                        "mappingType": 1,
                        "pattern": "src_as",
                        "thresholds": [],
                        "type": "hidden",
                        "unit": "short",
                    },
                    {
                        "alias": "",
                        "align": "auto",
                        "colorMode": None,
                        "colors": [
                            "rgba(245, 54, 54, 0.9)",
                            "rgba(237, 129, 40, 0.89)",
                            "rgba(50, 172, 45, 0.97)",
                        ],
                        "dateFormat": "YYYY-MM-DD HH:mm:ss",
                        "decimals": 2,
                        "mappingType": 1,
                        "pattern": "src_as_name",
                        "thresholds": [],
                        "type": "hidden",
                        "unit": "short",
                    },
                    {
                        "alias": "",
                        "align": "auto",
                        "colorMode": None,
                        "colors": [
                            "rgba(245, 54, 54, 0.9)",
                            "rgba(237, 129, 40, 0.89)",
                            "rgba(50, 172, 45, 0.97)",
                        ],
                        "dateFormat": "YYYY-MM-DD HH:mm:ss",
                        "decimals": 2,
                        "mappingType": 1,
                        "pattern": "instance",
                        "thresholds": [],
                        "type": "hidden",
                        "unit": "short",
                    },
                    {
                        "alias": "",
                        "align": "auto",
                        "colorMode": None,
                        "colors": [
                            "rgba(245, 54, 54, 0.9)",
                            "rgba(237, 129, 40, 0.89)",
                            "rgba(50, 172, 45, 0.97)",
                        ],
                        "dateFormat": "YYYY-MM-DD HH:mm:ss",
                        "decimals": 2,
                        "mappingType": 1,
                        "pattern": "job",
                        "thresholds": [],
                        "type": "hidden",
                        "unit": "short",
                    },
                    {
                        "alias": "ASN",
                        "align": "auto",
                        "colorMode": None,
                        "colors": [
                            "rgba(245, 54, 54, 0.9)",
                            "rgba(237, 129, 40, 0.89)",
                            "rgba(50, 172, 45, 0.97)",
                        ],
                        "dateFormat": "YYYY-MM-DD HH:mm:ss",
                        "decimals": 0,
                        "link": True,
                        "linkTargetBlank": True,
                        "linkTooltip": "PeeringDB",
                        "linkUrl": "https://as${__cell}.peeringdb.com",
                        "mappingType": 1,
                        "pattern": "src_as",
                        "thresholds": [],
                        "type": "number",
                        "unit": "none",
                    },
                    {
                        "alias": "AS Name",
                        "align": "auto",
                        "colorMode": None,
                        "colors": [
                            "rgba(245, 54, 54, 0.9)",
                            "rgba(237, 129, 40, 0.89)",
                            "rgba(50, 172, 45, 0.97)",
                        ],
                        "dateFormat": "YYYY-MM-DD HH:mm:ss",
                        "decimals": 2,
                        "mappingType": 1,
                        "pattern": "src_as_name",
                        "thresholds": [],
                        "type": "string",
                        "unit": "short",
                    },
                    {
                        "alias": "",
                        "align": "auto",
                        "colorMode": None,
                        "colors": [
                            "rgba(245, 54, 54, 0.9)",
                            "rgba(237, 129, 40, 0.89)",
                            "rgba(50, 172, 45, 0.97)",
                        ],
                        "dateFormat": "YYYY-MM-DD HH:mm:ss",
                        "decimals": 2,
                        "mappingType": 1,
                        "pattern": "hostname",
                        "thresholds": [],
                        "type": "hidden",
                        "unit": "short",
                    },
                ],
                "targets": [
                    {
                        "expr": 'topk(10, sum(rate(proxy_downstream_total{job="vipen-server",src_as=~"$as",instance=~"$server"}[$__range]) * 8) by (src_as, src_as_name))',
                        "format": "table",
                        "instant": True,
                        "legendFormat": "{{src_as}}",
                        "refId": "A",
                    }
                ],
                "timeFrom": None,
                "timeShift": None,
                "title": "Top Downstream ASNs",
                "transform": "table",
                "type": "table-old",
            },
            {
                "cacheTimeout": None,
                "columns": [],
                "datasource": "Prometheus",
                "fontSize": "100%",
                "gridPos": {"h": 11, "w": 12, "x": 12, "y": 12},
                "id": 6,
                "links": [],
                "pageSize": None,
                "pluginVersion": "6.3.6",
                "scroll": True,
                "showHeader": True,
                "sort": {"col": 8, "desc": True},
                "styles": [
                    {
                        "alias": "",
                        "align": "auto",
                        "colorMode": None,
                        "colors": [
                            "rgba(245, 54, 54, 0.9)",
                            "rgba(237, 129, 40, 0.89)",
                            "rgba(50, 172, 45, 0.97)",
                        ],
                        "dateFormat": "YYYY-MM-DD HH:mm:ss",
                        "decimals": 2,
                        "mappingType": 1,
                        "pattern": "Time",
                        "thresholds": [],
                        "type": "hidden",
                        "unit": "short",
                    },
                    {
                        "alias": "Bandwidth",
                        "align": "auto",
                        "colorMode": None,
                        "colors": [
                            "rgba(245, 54, 54, 0.9)",
                            "rgba(237, 129, 40, 0.89)",
                            "rgba(50, 172, 45, 0.97)",
                        ],
                        "dateFormat": "YYYY-MM-DD HH:mm:ss",
                        "decimals": 2,
                        "mappingType": 1,
                        "pattern": "Value",
                        "thresholds": [],
                        "type": "number",
                        "unit": "bps",
                    },
                    {
                        "alias": "",
                        "align": "auto",
                        "colorMode": None,
                        "colors": [
                            "rgba(245, 54, 54, 0.9)",
                            "rgba(237, 129, 40, 0.89)",
                            "rgba(50, 172, 45, 0.97)",
                        ],
                        "dateFormat": "YYYY-MM-DD HH:mm:ss",
                        "decimals": 2,
                        "mappingType": 1,
                        "pattern": "__name__",
                        "thresholds": [],
                        "type": "hidden",
                        "unit": "short",
                    },
                    {
                        "alias": "",
                        "align": "auto",
                        "colorMode": None,
                        "colors": [
                            "rgba(245, 54, 54, 0.9)",
                            "rgba(237, 129, 40, 0.89)",
                            "rgba(50, 172, 45, 0.97)",
                        ],
                        "dateFormat": "YYYY-MM-DD HH:mm:ss",
                        "decimals": 2,
                        "mappingType": 1,
                        "pattern": "src_as",
                        "thresholds": [],
                        "type": "hidden",
                        "unit": "short",
                    },
                    {
                        "alias": "",
                        "align": "auto",
                        "colorMode": None,
                        "colors": [
                            "rgba(245, 54, 54, 0.9)",
                            "rgba(237, 129, 40, 0.89)",
                            "rgba(50, 172, 45, 0.97)",
                        ],
                        "dateFormat": "YYYY-MM-DD HH:mm:ss",
                        "decimals": 2,
                        "mappingType": 1,
                        "pattern": "src_as_name",
                        "thresholds": [],
                        "type": "hidden",
                        "unit": "short",
                    },
                    {
                        "alias": "",
                        "align": "auto",
                        "colorMode": None,
                        "colors": [
                            "rgba(245, 54, 54, 0.9)",
                            "rgba(237, 129, 40, 0.89)",
                            "rgba(50, 172, 45, 0.97)",
                        ],
                        "dateFormat": "YYYY-MM-DD HH:mm:ss",
                        "decimals": 2,
                        "mappingType": 1,
                        "pattern": "instance",
                        "thresholds": [],
                        "type": "hidden",
                        "unit": "short",
                    },
                    {
                        "alias": "",
                        "align": "auto",
                        "colorMode": None,
                        "colors": [
                            "rgba(245, 54, 54, 0.9)",
                            "rgba(237, 129, 40, 0.89)",
                            "rgba(50, 172, 45, 0.97)",
                        ],
                        "dateFormat": "YYYY-MM-DD HH:mm:ss",
                        "decimals": 2,
                        "mappingType": 1,
                        "pattern": "job",
                        "thresholds": [],
                        "type": "hidden",
                        "unit": "short",
                    },
                    {
                        "alias": "ASN",
                        "align": "auto",
                        "colorMode": None,
                        "colors": [
                            "rgba(245, 54, 54, 0.9)",
                            "rgba(237, 129, 40, 0.89)",
                            "rgba(50, 172, 45, 0.97)",
                        ],
                        "dateFormat": "YYYY-MM-DD HH:mm:ss",
                        "decimals": 0,
                        "link": True,
                        "linkTargetBlank": True,
                        "linkTooltip": "PeeringDB",
                        "linkUrl": "https://as${__cell}.peeringdb.com",
                        "mappingType": 1,
                        "pattern": "src_as",
                        "thresholds": [],
                        "type": "number",
                        "unit": "none",
                    },
                    {
                        "alias": "AS Name",
                        "align": "auto",
                        "colorMode": None,
                        "colors": [
                            "rgba(245, 54, 54, 0.9)",
                            "rgba(237, 129, 40, 0.89)",
                            "rgba(50, 172, 45, 0.97)",
                        ],
                        "dateFormat": "YYYY-MM-DD HH:mm:ss",
                        "decimals": 2,
                        "mappingType": 1,
                        "pattern": "src_as_name",
                        "thresholds": [],
                        "type": "string",
                        "unit": "short",
                    },
                    {
                        "alias": "",
                        "align": "auto",
                        "colorMode": None,
                        "colors": [
                            "rgba(245, 54, 54, 0.9)",
                            "rgba(237, 129, 40, 0.89)",
                            "rgba(50, 172, 45, 0.97)",
                        ],
                        "dateFormat": "YYYY-MM-DD HH:mm:ss",
                        "decimals": 2,
                        "mappingType": 1,
                        "pattern": "hostname",
                        "thresholds": [],
                        "type": "hidden",
                        "unit": "short",
                    },
                ],
                "targets": [
                    {
                        "expr": 'topk(10, sum(rate(proxy_upstream_total{job="vipen-server",src_as=~"$as",instance=~"$server"}[$__range]) * 8) by (src_as, src_as_name))',
                        "format": "table",
                        "instant": True,
                        "legendFormat": "",
                        "refId": "A",
                    }
                ],
                "timeFrom": None,
                "timeShift": None,
                "title": "Top Upstream ASNs",
                "transform": "table",
                "type": "table-old",
            },
        ],
        "refresh": "30s",
        "schemaVersion": 32,
        "style": "dark",
        "tags": [],
        "templating": {
            "list": [
                {
                    "allValue": ".*",
                    "current": {"selected": False, "text": "All", "value": "$__all"},
                    "datasource": "Prometheus",
                    "definition": 'label_values(proxy_upstream_total{job="vipen-server"}, src_as)',
                    "description": None,
                    "error": None,
                    "hide": 0,
                    "includeAll": True,
                    "label": "ASN",
                    "multi": False,
                    "name": "as",
                    "options": [],
                    "query": {
                        "query": 'label_values(proxy_upstream_total{job="vipen-server"}, src_as)',
                        "refId": "Prometheus-as-Variable-Query",
                    },
                    "refresh": 1,
                    "regex": "",
                    "skipUrlSync": False,
                    "sort": 0,
                    "tagValuesQuery": "",
                    "tagsQuery": "",
                    "type": "query",
                    "useTags": False,
                },
                {
                    "allValue": ".*",
                    "current": {"selected": False, "text": "All", "value": "$__all"},
                    "datasource": "Prometheus",
                    "definition": 'label_values(proxy_downstream_total{job="vipen-server"}, cluster)',
                    "description": None,
                    "error": None,
                    "hide": 0,
                    "includeAll": True,
                    "label": "Cluster",
                    "multi": False,
                    "name": "cluster",
                    "options": [],
                    "query": {
                        "query": 'label_values(proxy_downstream_total{job="vipen-server"}, cluster)',
                        "refId": "Prometheus-host-Variable-Query",
                    },
                    "refresh": 1,
                    "regex": "",
                    "skipUrlSync": False,
                    "sort": 0,
                    "tagValuesQuery": "",
                    "tagsQuery": "",
                    "type": "query",
                    "useTags": False,
                },
                {
                    "allValue": ".*",
                    "current": {"selected": False, "text": "All", "value": "$__all"},
                    "datasource": "Prometheus",
                    "definition": 'label_values(proxy_downstream_total{job="vipen-server",cluster="$cluster"}, instance)',
                    "description": None,
                    "error": None,
                    "hide": 0,
                    "includeAll": True,
                    "label": "Server",
                    "multi": False,
                    "name": "server",
                    "options": [],
                    "query": {
                        "query": 'label_values(proxy_downstream_total{job="vipen-server",cluster="$cluster"}, instance)',
                        "refId": "Prometheus-host-Variable-Query",
                    },
                    "refresh": 1,
                    "regex": "",
                    "skipUrlSync": False,
                    "sort": 0,
                    "tagValuesQuery": "",
                    "tagsQuery": "",
                    "type": "query",
                    "useTags": False,
                },
                {
                    "datasource": None,
                    "description": None,
                    "error": None,
                    "filters": [],
                    "hide": 0,
                    "label": "",
                    "name": "Filters",
                    "skipUrlSync": False,
                    "type": "adhoc",
                },
            ]
        },
        "time": {"from": "now-1h", "to": "now"},
        "timepicker": {
            "refresh_intervals": [
                "5s",
                "10s",
                "30s",
                "1m",
                "5m",
                "15m",
                "30m",
                "1h",
                "2h",
                "1d",
            ]
        },
        "timezone": "",
        "title": "Trafic By AS",
        "uid": "txUPy4Rnz",
        "version": 1,
    },
    indent=4,
)



import click
from InquirerPy import prompt
from vipen.core.logging import console
from vipen.core.task import TaskPool

from .main import cli
from .util import current_cluster, get_config, new_progress


@cli.command()
@click.pass_context
def fix(ctx):
    config = get_config(ctx)
    cluster = current_cluster(ctx)

    # Find stale servers (servers that exist in provider but not in our)
    stale_servers = []
    for provider in config.server_providers:
        for server in provider.fetch_cluster_servers(cluster):
            if server not in cluster.servers:
                stale_servers.append(server)
    if len(stale_servers) > 0:
        console.log(f"Found {len(stale_servers)} stale servers:")
        for s in stale_servers:
            console.log(f"{s.provider.name}/{s.name}")

        answer = prompt(
            {
                "message": "Should I destroy the servers?",
                "type": "confirm",
                "default": True,
            },
        )
        if answer[0]:
            with new_progress() as p:
                with TaskPool(progress=p) as pool:
                    for s in stale_servers:
                        pool.submit(s.pname, s.destroy())

    # Find missing servers
    missing_servers = []
    for server in cluster.servers:
        if server.provider.fetch_server(server.id) is None:
            missing_servers.append(server)
    if len(missing_servers) > 0:
        console.log(f"Found {len(stale_servers)} missing servers:")
        for s in missing_servers:
            console.log(s.pname)

        answer = prompt(
            {
                "message": "Should I remove them?",
                "type": "confirm",
                "default": True,
            },
        )
        if answer[0]:
            for server in missing_servers:
                cluster.servers.remove(server)

    # Updating DNS Records
    def provision(domain):
        yield from domain.provision()
        yield from domain.update_dns_with_servers(cluster.servers)
        return True

    with new_progress() as p:
        with TaskPool(progress=p) as pool:
            for domain in cluster.domains:
                pool.submit(
                    f"{domain.name}",
                    provision(domain),
                )



import click
from rich import box
from rich.table import Table
from vipen.core.logging import console

from .main import cli
from .util import get_cluster


@cli.command(name="keys")
@click.option("--cluster", "-c", "cluster_name", default=None)
@click.pass_context
def keys(ctx, cluster_name):
    cluster = get_cluster(ctx, cluster_name)

    for domain in cluster.domains:
        console.print(f"[bold cyan]{domain.name}:[/bold cyan]")
        print(cluster.template.generate_key(cluster, domain))
        console.print("")



import traceback
from ast import List
from concurrent.futures import ThreadPoolExecutor
from email.generator import Generator

from rich.progress import Progress, SpinnerColumn, TaskID, TextColumn


class BaseError(Exception):
    pass


def new_progress():
    from vipen.core import console

    return Progress(
        SpinnerColumn(),
        TextColumn(
            "[blue]{task.description}:[/blue] {task.fields[message]}",
            justify="left",
        ),
        console=console,
    )


def run_task(description: str, g: Generator, progress: Progress, raise_exc=False):
    task_id = progress.add_task(description, message="", start=True, total=1)
    last_message = ""
    try:
        while True:
            step = next(g)
            last_message = step
            progress.update(task_id, message=step)
    except StopIteration as e:
        progress.update(task_id, message=f"{last_message} [green]OK")
        progress.advance(task_id, 1)
        return e.value
    except BaseError as e:
        progress.update(
            task_id,
            message=f"{last_message} [red]Fail[/red] [deep_pink4]{str(e)}[/deep_pink4]",
        )
        progress.advance(task_id, 1)
        if raise_exc:
            raise
        return None
    except Exception as e:
        progress.console.error(traceback.format_exc())
        progress.update(
            task_id,
            message=f"{last_message} [red]Fail[/red] [deep_pink4]{str(e)}[/deep_pink4]",
        )
        progress.advance(task_id, 1)
        if raise_exc:
            raise
        return None


def with_prefix(prefix: str, g: Generator, style="light_slate_grey") -> Generator:
    yield prefix
    for m in g:
        if style:
            yield f"{prefix} [{style}]{m}[/{style}]"
        else:
            yield f"{prefix} {m}"


class TaskPool(ThreadPoolExecutor):
    progress: Progress
    _futures: List

    def __init__(self, progress: Progress, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.progress = progress
        self._futures = []

    def submit(self, description: str, t: Generator):
        r = super().submit(
            run_task, description, t, progress=self.progress, raise_exc=True
        )
        self._futures.append(r)
        return r

    @property
    def results(self):
        return [f.result() for f in self._futures]

    @property
    def exceptions(self):
        return [f.exception for f in self._futures]



import time
from typing import Generator, List

import paramiko
import tld
from vipen.core.logging import console
from vipen.core.task import BaseError
from vipen.core.utils import PromptField, dataclass


class ProviderError(BaseError):
    pass


class ProvisioningError(BaseError):
    pass


@dataclass
class Provider:
    domain_cls = None
    server_cls = None

    cloud: str
    name: str

    @property
    def can_server(self):
        return self.server_cls is not None

    @property
    def can_cdn(self):
        return self.domain_cls is not None

    def init_provider(self):
        yield "Provider ready"


@dataclass
class ServerProvider(Provider):
    ssh_key: str = PromptField(default="~/.ssh/id_rsa", type="filepath")
    ssh_user: str = PromptField(default="root")

    def create_server(self, name) -> "Server":
        raise NotImplemented()

    def fetch_cluster_servers(self, cluster) -> List["Server"]:
        raise NotImplemented()

    def fetch_server(self, id) -> "Server":
        raise NotImplemented()


@dataclass
class CDNProvider(Provider):
    def create_domain(self, domain_name: str) -> Generator:
        raise NotImplemented()


@dataclass
class Server:
    name: str
    ip: str
    provider: Provider
    id: str

    _client: paramiko.SSHClient = None

    @property
    def pname(self) -> str:
        return f"{self.provider.name}/{self.name}"

    @property
    def is_connected(self):
        return self._client is not None

    def connect(self, retries=5, retry_backoff=4) -> Generator:
        for r in range(retries):
            yield f"attempt [{r+1}/{retries}]"
            client = paramiko.SSHClient()
            client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            user = self.provider.ssh_user or "root"
            try:
                client.connect(
                    self.ip, port=22, username=user, key_filename=self.provider.ssh_key
                )
                self._client = client
                return
            except Exception as e:
                if r == retries - 1:
                    self._client = None
                    raise
                time.sleep(retry_backoff)

    def run(self, *cmd) -> Generator:
        if self._client is None:
            raise ProviderError("Client not connected")

        tran = self._client.get_transport()
        chan = tran.open_session()
        chan.get_pty()
        f = chan.makefile()
        chan.exec_command(" ".join(cmd))

        lines = []
        max_line_len = 80
        while True:
            line = f.readline()
            if not line:
                break
            line = line.strip()
            lines.append(line)
            if len(line) > max_line_len:
                yield line[:max_line_len] + "..."
            else:
                yield line

        if chan.recv_exit_status() != 0:
            console.error(lines[-15:])
            raise ProvisioningError()

    def destroy(self) -> Generator:
        raise NotImplemented()


@dataclass
class Domain:
    name: str
    provider: Provider

    @property
    def sld(self):
        """
        Second Level Domain (e.g, internal.chat.google.com -> google.com)
        """
        p = tld.parse_tld(self.name, fix_protocol=True)
        return f"{p[1]}.{p[0]}"

    @property
    def dns_name(self):
        """
        After SLD (e.g, internal.chat.google.com -> internal.chat)
        """
        p = tld.parse_tld(self.name, fix_protocol=True)
        p = [x for x in p if x]
        if len(p) == 2:
            return "@"
        return ".".join(p[2:][::-1])

    def update_dns_with_servers(
        self, servers: List[Server], cloud_proxy=True
    ) -> Generator:
        raise NotImplemented()

    def is_active(self, reload=False) -> bool:
        raise NotImplemented()

    def ns_servers(self, reload=False) -> List[str]:
        raise NotImplemented()

    def provision(self) -> Generator:
        raise NotImplemented()



import click
from InquirerPy import prompt
from InquirerPy.base.control import Choice
from rich import box
from rich.table import Table
from vipen.core.logging import console
from vipen.core.task import run_task
from vipen.core.utils import prompt_factory

from .main import cli
from .util import get_config, new_progress


@cli.group()
def provider():
    pass


@provider.command(name="add")
@click.argument("provider_name", nargs=1, default="")
@click.pass_context
def add_provider(ctx, provider_name):
    config = get_config(ctx)

    answer = prompt(
        {
            "message": "Select Provider Cloud",
            "type": "rawlist",
            "choices": [
                Choice(name=f"{p.cloud}", value=p) for p in config.provider_cls
            ],
        },
    )
    provider_cls = answer[0]

    if not provider_name:
        answer = prompt(
            {
                "message": "Enter Provider Name",
                "type": "input",
                "validate": lambda v: v and " " not in v,
            },
        )
        provider_name = answer[0]
    if config.get_provider(provider_name) is not None:
        console.error(f"Provider with name {provider_name} already exists")
        return

    params = prompt(prompt_factory(provider_cls))
    provider = provider_cls(cloud=provider_cls.cloud, name=provider_name, **params)

    with new_progress() as p:
        run_task(provider.name, provider.init_provider(), progress=p)

    config.providers.append(provider)

    console.log(f"New provider '{provider_name}' created")


@provider.command(name="rm")
@click.argument(
    "provider_name",
    nargs=1,
)
@click.pass_context
def remove_provider(ctx, provider_name):
    config = get_config(ctx)

    provider = config.get_provider(provider_name)
    if provider is None:
        console.error(f"Provider {provider_name} not found")
        return

    config.providers.remove(provider)
    console.log(f"Provider '{provider_name}' removed")


@provider.command(name="ls")
@click.option("--show-api-key", "-a", type=bool, is_flag=True, default=False)
@click.pass_context
def list_providers(ctx, show_api_key):
    config = get_config(ctx)

    table = Table(box=box.SIMPLE_HEAVY, show_header=True, show_footer=False)

    table.add_column("Cloud", justify="right", style="cyan", no_wrap=True)
    table.add_column("Name", style="magenta")
    table.add_column("Supports Servers", style="magenta")
    table.add_column("Supports Domains", style="magenta")
    if show_api_key:
        table.add_column("API Key", style="magenta")

    bool_str = lambda b: "yes" if b else "no"
    for p in config.providers:
        f = [p.cloud, p.name, bool_str(p.can_server), bool_str(p.can_cdn)]
        if show_api_key:
            f.append(p.api_key)
        table.add_row(*f)

    console.print(table)



import click
from InquirerPy import prompt
from InquirerPy.base.control import Choice
from rich import box
from rich.table import Table
from vipen.core.config import Cluster
from vipen.core.logging import console
from vipen.core.utils import prompt_factory

from .main import cli
from .util import get_config


@cli.group()
def cluster():
    pass


@cluster.command(name="add")
@click.argument("cluster_name", nargs=1, default="")
@click.pass_context
def add_cluster(ctx, cluster_name):
    config = get_config(ctx)

    if not cluster_name:
        answer = prompt(
            {
                "message": "Enter Cluster Name",
                "type": "input",
                "validate": lambda v: v and " " not in v,
            },
        )
        cluster_name = answer[0]

    if config.get_cluster(cluster_name) is not None:
        console.error(f"Cluster with name '{cluster_name}' already exists")
        return

    cluster = Cluster.new(cluster_name)

    answer = prompt(
        {
            "message": "Select Template to use in Cluster",
            "type": "rawlist",
            "choices": [
                Choice(name=f"{t.description}", value=t) for t in config.template_cls
            ],
        },
    )
    template_cls = answer[0]

    params = prompt(prompt_factory(template_cls))
    template = template_cls(type=template_cls.type, **params)
    cluster.template = template

    config.clusters.append(cluster)
    console.log(f"Created new cluster '{cluster_name}'")

    if not config.current_cluster:
        config.current_cluster = cluster
        console.log(f"Set '{cluster_name}' as default cluster")
    else:
        console.log(f"To use this cluster by default run <TODO>")


@cluster.command(name="rm")
@click.argument(
    "cluster_name",
    nargs=1,
)
@click.pass_context
def remove_cluster(ctx, cluster_name):
    config = get_config(ctx)

    cluster = config.get_cluster(cluster_name)
    if cluster is None:
        console.error(f"Cluster {cluster_name} not found")
        return

    console.log("Removing cluster will not destroy it's resources.")
    console.log("Make sure you scale down the cluster before continuing.")

    answer = prompt(
        {
            "message": "Do you wish to continue?",
            "type": "confirm",
            "default": False,
        },
    )
    if not answer[0]:
        return

    config.clusters.remove(cluster)
    console.log(f"Cluster '{cluster_name}' removed")


@cluster.command(name="ls")
@click.pass_context
def list_clusters(ctx):
    config = get_config(ctx)

    table = Table(box=box.SIMPLE_HEAVY, show_header=True, show_footer=False)

    table.add_column("Name", justify="right", style="cyan", no_wrap=True)
    table.add_column("Template", justify="right", style="cyan", no_wrap=True)
    table.add_column("Server Count", style="magenta")
    table.add_column("Domain Count", style="magenta")

    for c in config.clusters:
        table.add_row(c.name, c.template.type, str(len(c.servers)), str(len(c.domains)))

    console.print(table)


@cluster.command(name="use")
@click.argument("cluster_name", nargs=1, default="")
@click.pass_context
def use_cluster(ctx, cluster_name):
    config = get_config(ctx)

    if not cluster_name:
        answer = prompt(
            {
                "message": "Choose Cluster",
                "type": "rawlist",
                "choices": [c.name for c in config.clusters],
            },
        )
        cluster_name = answer[0]

    cluster = config.get_cluster(cluster_name)
    if cluster is None:
        console.error(f"Cluster {cluster_name} not found")
        return

    config.current_cluster = cluster
    console.log(f"Set '{cluster.name}' as default cluster")



import dataclasses
import os

import click
from InquirerPy import prompt
from InquirerPy.base.control import Choice
from vipen.core.config import Cluster, Config, Monitoring
from vipen.core.logging import console
from vipen.core.task import TaskPool, run_task, with_prefix
from vipen.templates.base import script_install_docker, script_install_docker_compose

from .main import cli
from .util import choose_server_provider, current_cluster, get_config, new_progress


@cli.group()
def monitoring():
    pass


@monitoring.command(name="init")
@click.option("--provider", "-p", "provider_name", default=None)
@click.pass_context
def init_monitoring(ctx, provider_name):
    config: Config = get_config(ctx)
    provider = choose_server_provider(ctx, provider_name)

    if config.monitoring is not None:
        console.error("Monitoring has already been initialized.")
        console.error(
            "Consider destroying the current configuration if you want to reinitialize."
        )
        return

    candidate_domains = [
        dataclasses.replace(d, name=f"monitoring.{d.sld}")
        for d in config.get_all_domains()
    ]

    answer = prompt(
        {
            "message": "Which domain do you want to use?",
            "type": "rawlist",
            "choices": [
                Choice(name=candidate_domains[d].name, value=d)
                for d in range(len(candidate_domains))
            ],
        },
    )
    domain = candidate_domains[answer[0]]

    with new_progress() as p:
        server = run_task(
            "Monitoring Server",
            provider.create_server(Cluster.new("monitoring"), "monitoring"),
            progress=p,
        )

    from vipen.providers.digitalocean import DigitalOceanServer

    server = DigitalOceanServer(
        name="monitoring",
        ip="159.89.28.107",
        id="325969881",
        provider=config.server_providers[0],
    )
    if server is None:
        return

    with new_progress() as p:
        run_task(
            domain.name,
            domain.update_dns_with_servers([server], cloud_proxy=False),
            progress=p,
        )

    monitoring = Monitoring(server=server, domain=domain)
    with new_progress() as p:
        run_task(
            server.name,
            monitoring.provision_server(config),
            progress=p,
        )

    config.monitoring = monitoring


@monitoring.command(name="destroy")
@click.pass_context
def destroy_monitoring(ctx):
    config = get_config(ctx)

    if config.monitoring is None:
        console.log("No monitoring configration exists")
        return

    server = config.monitoring.server
    domain = config.monitoring.domain
    with new_progress() as p:
        with TaskPool(progress=p) as pool:
            if server is not None:
                pool.submit(server.pname, server.destroy())
            if domain is not None:
                pool.submit(domain.name, domain.update_dns_with_servers([]))



import json

dashboard = json.dumps(
    {
        "annotations": {
            "list": [
                {
                    "builtIn": 1,
                    "datasource": {"type": "grafana", "uid": "-- Grafana --"},
                    "enable": True,
                    "hide": True,
                    "iconColor": "rgba(0, 211, 255, 1)",
                    "name": "Annotations & Alerts",
                    "target": {
                        "limit": 100,
                        "matchAny": False,
                        "tags": [],
                        "type": "dashboard",
                    },
                    "type": "dashboard",
                }
            ]
        },
        "editable": True,
        "fiscalYearStartMonth": 0,
        "graphTooltip": 0,
        "id": 4,
        "links": [],
        "liveNow": False,
        "panels": [
            {
                "datasource": "Prometheus",
                "description": "",
                "fieldConfig": {
                    "defaults": {
                        "color": {"mode": "thresholds"},
                        "mappings": [],
                        "thresholds": {
                            "mode": "absolute",
                            "steps": [{"color": "semi-dark-green", "value": None}],
                        },
                        "unit": "Bps",
                    },
                    "overrides": [],
                },
                "gridPos": {"h": 4, "w": 5, "x": 0, "y": 0},
                "id": 16,
                "options": {
                    "colorMode": "background",
                    "graphMode": "area",
                    "justifyMode": "auto",
                    "orientation": "auto",
                    "reduceOptions": {
                        "calcs": ["lastNotNull"],
                        "fields": "",
                        "values": False,
                    },
                    "textMode": "value",
                },
                "pluginVersion": "9.1.7",
                "targets": [
                    {
                        "expr": 'sum(rate(proxy_downstream_total{job="vipen-server"}[5m]))',
                        "legendFormat": "__auto",
                        "range": True,
                        "refId": "A",
                    }
                ],
                "title": "Global Downstream",
                "type": "stat",
            },
            {
                "datasource": "Prometheus",
                "description": "",
                "fieldConfig": {
                    "defaults": {
                        "color": {"mode": "thresholds"},
                        "mappings": [],
                        "thresholds": {
                            "mode": "absolute",
                            "steps": [{"color": "semi-dark-blue", "value": None}],
                        },
                        "unit": "Bps",
                    },
                    "overrides": [],
                },
                "gridPos": {"h": 4, "w": 5, "x": 5, "y": 0},
                "id": 18,
                "options": {
                    "colorMode": "background",
                    "graphMode": "area",
                    "justifyMode": "auto",
                    "orientation": "auto",
                    "reduceOptions": {
                        "calcs": ["lastNotNull"],
                        "fields": "",
                        "values": False,
                    },
                    "textMode": "value",
                },
                "pluginVersion": "9.1.7",
                "targets": [
                    {
                        "expr": 'sum(rate(proxy_upstream_total{job="vipen-server"}[5m]))',
                        "legendFormat": "__auto",
                        "range": True,
                        "refId": "A",
                    }
                ],
                "title": "Global Upstream",
                "type": "stat",
            },
            {
                "datasource": "Prometheus",
                "description": "",
                "fieldConfig": {
                    "defaults": {
                        "color": {"mode": "thresholds"},
                        "mappings": [],
                        "thresholds": {
                            "mode": "absolute",
                            "steps": [{"color": "semi-dark-purple", "value": None}],
                        },
                        "unit": "reqps",
                    },
                    "overrides": [],
                },
                "gridPos": {"h": 4, "w": 5, "x": 10, "y": 0},
                "id": 20,
                "options": {
                    "colorMode": "background",
                    "graphMode": "area",
                    "justifyMode": "auto",
                    "orientation": "auto",
                    "reduceOptions": {
                        "calcs": ["lastNotNull"],
                        "fields": "",
                        "values": False,
                    },
                    "textMode": "value",
                },
                "pluginVersion": "9.1.7",
                "targets": [
                    {
                        "expr": 'sum(rate(proxy_requests_total{job="vipen-server"}[5m]))',
                        "legendFormat": "__auto",
                        "range": True,
                        "refId": "A",
                    }
                ],
                "title": "Global Requests",
                "type": "stat",
            },
            {
                "datasource": "Prometheus",
                "description": "",
                "fieldConfig": {
                    "defaults": {
                        "color": {"mode": "thresholds"},
                        "mappings": [],
                        "thresholds": {
                            "mode": "absolute",
                            "steps": [{"color": "semi-dark-orange", "value": None}],
                        },
                        "unit": "decbytes",
                    },
                    "overrides": [],
                },
                "gridPos": {"h": 4, "w": 5, "x": 15, "y": 0},
                "id": 22,
                "options": {
                    "colorMode": "background",
                    "graphMode": "area",
                    "justifyMode": "auto",
                    "orientation": "auto",
                    "reduceOptions": {
                        "calcs": ["lastNotNull"],
                        "fields": "",
                        "values": False,
                    },
                    "textMode": "value",
                },
                "pluginVersion": "9.1.7",
                "targets": [
                    {
                        "expr": 'sum(increase(proxy_downstream_total{job="vipen-server"}[24h])) + sum(increase(proxy_upstream_total{job="vipen-server"}[24h]))',
                        "legendFormat": "__auto",
                        "range": True,
                        "refId": "A",
                    }
                ],
                "title": "Global Daily Traffic",
                "type": "stat",
            },
            {
                "datasource": "Prometheus",
                "description": "",
                "fieldConfig": {
                    "defaults": {
                        "color": {"mode": "thresholds"},
                        "mappings": [],
                        "thresholds": {
                            "mode": "absolute",
                            "steps": [{"color": "semi-dark-green", "value": None}],
                        },
                        "unit": "Bps",
                    },
                    "overrides": [],
                },
                "gridPos": {"h": 4, "w": 5, "x": 0, "y": 4},
                "id": 4,
                "options": {
                    "colorMode": "none",
                    "graphMode": "area",
                    "justifyMode": "auto",
                    "orientation": "auto",
                    "reduceOptions": {
                        "calcs": ["lastNotNull"],
                        "fields": "",
                        "values": False,
                    },
                    "textMode": "value",
                },
                "pluginVersion": "9.1.7",
                "repeat": "cluster",
                "repeatDirection": "v",
                "targets": [
                    {
                        "expr": 'sum(rate(proxy_downstream_total{job="vipen-server",cluster="$cluster"}[5m]))',
                        "legendFormat": "__auto",
                        "range": True,
                        "refId": "A",
                    }
                ],
                "title": "$cluster - Downstream",
                "type": "stat",
            },
            {
                "datasource": "Prometheus",
                "datasource": "Prometheus",
                "description": "",
                "fieldConfig": {
                    "defaults": {
                        "color": {"mode": "thresholds"},
                        "mappings": [],
                        "thresholds": {
                            "mode": "absolute",
                            "steps": [{"color": "semi-dark-blue", "value": None}],
                        },
                        "unit": "Bps",
                    },
                    "overrides": [],
                },
                "gridPos": {"h": 4, "w": 5, "x": 5, "y": 4},
                "id": 6,
                "options": {
                    "colorMode": "none",
                    "graphMode": "area",
                    "justifyMode": "auto",
                    "orientation": "auto",
                    "reduceOptions": {
                        "calcs": ["lastNotNull"],
                        "fields": "",
                        "values": False,
                    },
                    "textMode": "value",
                },
                "pluginVersion": "9.1.7",
                "repeat": "cluster",
                "repeatDirection": "v",
                "targets": [
                    {
                        "expr": 'sum(rate(proxy_upstream_total{job="vipen-server",cluster="$cluster"}[5m]))',
                        "legendFormat": "__auto",
                        "range": True,
                        "refId": "A",
                    }
                ],
                "title": "$cluster - Upstream",
                "type": "stat",
            },
            {
                "datasource": "Prometheus",
                "description": "",
                "fieldConfig": {
                    "defaults": {
                        "color": {"mode": "thresholds"},
                        "mappings": [],
                        "thresholds": {
                            "mode": "absolute",
                            "steps": [{"color": "semi-dark-purple", "value": None}],
                        },
                        "unit": "reqps",
                    },
                    "overrides": [],
                },
                "gridPos": {"h": 4, "w": 5, "x": 10, "y": 4},
                "id": 8,
                "options": {
                    "colorMode": "none",
                    "graphMode": "area",
                    "justifyMode": "auto",
                    "orientation": "auto",
                    "reduceOptions": {
                        "calcs": ["lastNotNull"],
                        "fields": "",
                        "values": False,
                    },
                    "textMode": "value",
                },
                "pluginVersion": "9.1.7",
                "repeat": "cluster",
                "repeatDirection": "v",
                "targets": [
                    {
                        "expr": 'sum(rate(proxy_requests_total{job="vipen-server",cluster="$cluster"}[5m]))',
                        "legendFormat": "__auto",
                        "range": True,
                        "refId": "A",
                    }
                ],
                "title": "$cluster - Requests",
                "type": "stat",
            },
            {
                "datasource": "Prometheus",
                "description": "",
                "fieldConfig": {
                    "defaults": {
                        "color": {"mode": "thresholds"},
                        "mappings": [],
                        "thresholds": {
                            "mode": "absolute",
                            "steps": [{"color": "semi-dark-orange", "value": None}],
                        },
                        "unit": "decbytes",
                    },
                    "overrides": [],
                },
                "gridPos": {"h": 4, "w": 5, "x": 15, "y": 4},
                "id": 10,
                "options": {
                    "colorMode": "none",
                    "graphMode": "area",
                    "justifyMode": "auto",
                    "orientation": "auto",
                    "reduceOptions": {
                        "calcs": ["lastNotNull"],
                        "fields": "",
                        "values": False,
                    },
                    "textMode": "value",
                },
                "pluginVersion": "9.1.7",
                "repeat": "cluster",
                "repeatDirection": "v",
                "targets": [
                    {
                        "expr": 'sum(increase(proxy_downstream_total{job="vipen-server", cluster="$cluster"}[24h])) + sum(increase(proxy_upstream_total{job="vipen-server",cluster="$cluster"}[24h]))',
                        "legendFormat": "__auto",
                        "range": True,
                        "refId": "A",
                    }
                ],
                "title": "$cluster- Daily",
                "type": "stat",
            },
        ],
        "schemaVersion": 37,
        "style": "dark",
        "tags": [],
        "templating": {
            "list": [
                {
                    "current": {"selected": False, "text": "All", "value": "$__all"},
                    "datasource": "Prometheus",
                    "definition": "label_values(proxy_requests_total,cluster)",
                    "hide": 1,
                    "includeAll": True,
                    "label": "Cluster",
                    "multi": False,
                    "name": "cluster",
                    "options": [],
                    "query": {
                        "query": "label_values(proxy_requests_total,cluster)",
                        "refId": "StandardVariableQuery",
                    },
                    "refresh": 1,
                    "regex": "",
                    "skipUrlSync": False,
                    "sort": 1,
                    "type": "query",
                }
            ]
        },
        "time": {"from": "now-6h", "to": "now"},
        "timepicker": {},
        "timezone": "",
        "title": "Overview",
        "uid": "m5ZbtiS4k",
        "version": 8,
        "weekStart": "",
    },
    indent=4,
)



from click import ClickException, Context
from InquirerPy import prompt
from InquirerPy.base.control import Choice
from rich.progress import Progress, SpinnerColumn, TextColumn
from vipen.core.config import Cluster, Config
from vipen.core.logging import console
from vipen.providers.base import CDNProvider, ServerProvider


def new_progress():
    return Progress(
        SpinnerColumn(),
        TextColumn(
            "[blue]{task.description}:[/blue] {task.fields[message]}",
            justify="left",
        ),
        console=console,
    )


def get_config(ctx: Context) -> Config:
    return ctx.obj["config"]


def current_cluster(ctx: Context) -> Cluster:
    cfg = get_config(ctx)
    if cfg.current_cluster is None:
        raise ClickException("No cluster has been selected")
    return cfg.current_cluster


def get_cluster(ctx: Context, cluster_name: str = None) -> Cluster:
    cfg = get_config(ctx)
    if cluster_name is None:
        if cfg.current_cluster is None:
            raise ClickException("No cluster has been selected")
        return cfg.current_cluster
    cluster = cfg.clusters.get_cluster(cluster_name)
    if cluster is None:
        raise ClickException(f"Cluster '{cluster_name}' does not exist")
    return cluster


def choose_server_provider(ctx: Context, name=None) -> ServerProvider:
    cfg = get_config(ctx)

    if name is not None:
        if name not in cfg.providers:
            raise ClickException(f"Provider '{name}' does not exist")
        return cfg.providers[name]

    provider = None
    server_providers = cfg.server_providers
    if len(server_providers) == 0:
        raise ClickException(f"No server providers found")
    elif len(server_providers) == 1:
        provider = server_providers[0]
    else:
        answer = prompt(
            {
                "message": "Select server provider:",
                "type": "fuzzy",
                "choices": [
                    Choice(name=f"{p.cloud} - {p.name}", value=p)
                    for p in server_providers
                ],
            },
        )
        provider = answer[0]

    ctx.obj["server_provider"] = provider
    return provider


def choose_cdn_provider(ctx: Context, name=None) -> CDNProvider:
    cfg = get_config(ctx)

    if name is not None:
        if name not in cfg.providers:
            raise ClickException(f"Provider '{name}' does not exist")
        return cfg.providers[name]

    provider = None
    cdn_providers = cfg.cdn_providers
    if len(cdn_providers) == 0:
        raise ClickException(f"No CDN providers found")
    elif len(cdn_providers) == 1:
        provider = cdn_providers[0]
    else:
        answer = prompt(
            {
                "message": "Select CDN provider:",
                "type": "fuzzy",
                "choices": [
                    Choice(name=f"{p.cloud} - {p.name}", value=p) for p in cdn_providers
                ],
            },
        )
        provider = answer[0]

    ctx.obj["cdn_provider"] = provider
    return provider


def get_server_provider(ctx: Context) -> ServerProvider:
    provider = ctx.obj.get("server_provider", None)
    if provider is None:
        raise ClickException("Server provider not selected")
    return provider


def get_cdn_provider(ctx: Context) -> CDNProvider:
    provider = ctx.obj.get("cdn_provider", None)
    if provider is None:
        raise ClickException("CDN provider not selected")
    return provider



import base64
import json
from dataclasses import dataclass
from uuid import uuid4

from vipen.core.config import Cluster
from vipen.core.task import with_prefix
from vipen.core.utils import PromptField
from vipen.providers import Domain, Server

from .base import (
    Template,
    script_install_docker,
    script_install_docker_compose,
    tls_crt,
    tls_key,
)


@dataclass
class V2RayWSTlsNoAEAD(Template):
    type = "v2ray_ws_tls_noaead"
    description = "V2Ray Websocket+TLS (No AEAD)"

    id: str = PromptField(default_factory=lambda: str(uuid4()))

    vipen_dir = "/root/vipen"
    v2ray_path = "/vipen"

    def provision_server(self, server: Server):
        if not server.is_connected:
            yield from with_prefix("Connecting to server", server.connect())

        yield from with_prefix("Installing Docker", server.run(script_install_docker))

        yield from with_prefix(
            "Installing Docker Compose", server.run(script_install_docker_compose)
        )

        yield from with_prefix(
            "Running Services",
            server.run(
                f"""
                    mkdir -p {self.vipen_dir}
                    cd {self.vipen_dir}

                    mkdir cert
                    echo '{tls_crt}' > cert/tls.crt
                    echo '{tls_key}' > cert/tls.key

                    echo '{docker_compose}' > docker-compose.yml
                    echo '{v2ray_config(id=self.id, path=self.v2ray_path)}' > v2ray.json
                    echo '{nginx_config(path=self.v2ray_path)}' > nginx.conf
                    docker-compose up -d
                """,
            ),
        )

    def generate_key(self, cluster: Cluster, domain: Domain):
        key_data = {
            "add": domain.name,
            "aid": "0",
            "alpn": "",
            "host": domain.name,
            "id": self.id,
            "net": "ws",
            "path": self.v2ray_path,
            "port": "443",
            "ps": f"{cluster.name}-{domain.sld}",
            "scy": "none",
            "sni": domain.name,
            "tls": "tls",
            "type": "",
            "v": "2",
        }
        return f"vmess://{base64.b64encode(json.dumps(key_data).encode('utf-8')).decode('utf-8')}"


docker_compose = """
version: "3"
services:
  v2ray:
    image: v2fly/v2fly-core
    command: ["run", "-c", "/etc/v2ray/config.json"]
    restart: always
    environment:
      - V2RAY_VMESS_AEAD_FORCED=false
    volumes:
      - ./v2ray.json:/etc/v2ray/config.json:ro
  wsproxy:
    image: ghcr.io/ghitf/wsproxy:main
    command: ["--backend", "ws://v2ray:6000", "--metrics", "0.0.0.0:9900"]
    restart: always
  nginx:
    image: nginx
    restart: always
    ports:
      - 80:80
      - 443:443
    volumes:
      - ./cert:/root/cert
      - ./nginx.conf:/etc/nginx/conf.d/default.conf
""".strip()

v2ray_config = (
    lambda id, path: """
{
    "inbounds": [{
        "port": 6000,
        "listen": "0.0.0.0",
        "protocol": "vmess",
        "settings": {
            "clients": [{
                "id": "%s",
                "alterId": 64
            }]
        },
        "streamSettings": {
            "network": "ws",
            "wsSettings": {
                "path": "%s"
            }
        }
    }],
    "outbounds": [{
        "protocol": "freedom",
        "settings": {}
    }]
}
""".strip()
    % (id, path)
)

nginx_config = (
    lambda path: """
server {
    listen 443 ssl;
    ssl_certificate       /root/cert/tls.crt;
    ssl_certificate_key   /root/cert/tls.key;
    ssl_protocols         TLSv1 TLSv1.1 TLSv1.2;
    ssl_ciphers           HIGH:!aNULL:!MD5;

    location %s {
      if ($http_upgrade != "websocket") { # Return 404 error when WebSocket upgrading negotiate failed
          return 404;
      }
      proxy_redirect off;
      proxy_pass http://wsproxy;
      proxy_http_version 1.1;
      proxy_set_header Upgrade $http_upgrade;
      proxy_set_header Connection "upgrade";
      proxy_set_header Host $host;
      # Show real IP in v2ray access.log
      proxy_set_header X-Real-IP $remote_addr;
      proxy_set_header X-Forwarded-For $proxy_add_x_forwarded_for;
    }

    location /metrics {
        proxy_redirect off;
        proxy_pass http://wsproxy:9900;
        proxy_http_version 1.1;
    }

    location / {
        try_files $uri $uri/ =404;
    }
}

server {
    listen 80;

    location %s {
      if ($http_upgrade != "websocket") { # Return 404 error when WebSocket upgrading negotiate failed
          return 404;
      }
      proxy_redirect off;
      proxy_pass http://wsproxy;
      proxy_http_version 1.1;
      proxy_set_header Upgrade $http_upgrade;
      proxy_set_header Connection "upgrade";
      proxy_set_header Host $host;
      # Show real IP in v2ray access.log
      proxy_set_header X-Real-IP $remote_addr;
      proxy_set_header X-Forwarded-For $proxy_add_x_forwarded_for;
    }

    location /metrics {
        proxy_redirect off;
        proxy_pass http://wsproxy:9900;
        proxy_http_version 1.1;
    }

    location / {
        try_files $uri $uri/ =404;
    }
}
""".strip()
    % (path, path)
)



import os
import time
from email.generator import Generator
from json import JSONDecodeError
from typing import List

import requests
from vipen.core.logging import console
from vipen.core.task import BaseError
from vipen.core.utils import PromptField, dataclass

from .base import CDNProvider, Domain, ProviderError, ProvisioningError, Server


class APIError(BaseError):
    pass


class ArvanAPI:
    api_url: str = "https://napi.arvancloud.com/cdn/4.0/"

    def __init__(self, key):
        self.api_key = key

    def request(self, method, url_parts, *args, **kwargs):
        url_parts = [u.strip("/") for u in url_parts]
        url = os.path.join(self.api_url, *url_parts)
        resp = requests.request(
            method,
            url,
            headers={
                "Authorization": f"ApiKey {self.api_key}",
                "Accept": "application/json",
            },
            # proxies={"https": "http://localhost:8080"},
            # verify=False,
            # allow_redirects=False,
            *args,
            **kwargs,
        )
        return self.handle_data_response(resp)

    def get(self, *args):
        return self.request("GET", args)

    def delete(self, *args):
        return self.request("DELETE", args)

    def post(self, *args, json=None):
        return self.request("POST", args, json=json)

    def put(self, *args, json=None):
        return self.request("PUT", args, json=json)

    def patch(self, *args, json=None):
        return self.request("PATCH", args, json=json)

    def handle_data_response(self, resp: requests.Response):
        try:
            if resp.status_code == 404:
                return None
            if 200 <= resp.status_code < 300:
                return resp.json().get("data")
            if 300 <= resp.status_code < 400:
                raise APIError(
                    f"Recieved redirect response for '{resp.headers['Location']}'"
                )
            raise APIError(f"{resp.status_code}: {resp.json()['message']}")
        except JSONDecodeError:
            raise APIError(f"{resp.status_code}: {resp.text}")


@dataclass
class ArvanDomain(Domain):
    def update_dns_with_servers(
        self, servers: List[Server], cloud_proxy=True
    ) -> Generator:
        yield f"Fetching domain {self.sld}"

        api_domain = self.api.get("domains", self.sld)
        if api_domain is None:
            raise ProviderError(f"Domain {self.sld} does not exist in Arvan account")

        if api_domain.get("status", "") != "active":
            console.warn(f"Domain {self.sld} is not active")

        dns_update = {
            "type": "a",
            "name": self.dns_name,
            "value": [
                {"ip": s.ip, "port": None, "weight": 1, "country": ""} for s in servers
            ],
            "ttl": 120,
            "cloud": cloud_proxy,
            "upstream_https": "default",
            "ip_filter_mode": {"count": "multi", "order": "rr", "geo_filter": "none"},
        }

        yield f"Fetching DNS records for {self.sld}"
        dns_records = self.api.get("domains", self.sld, "dns-records")

        for record in dns_records:
            if record.get("type") == "a" and record.get("name") == self.dns_name:
                # Arvan errors on duplicates, so only send update if something changed
                needs_update = False
                dns_update["value"].sort(key=lambda v: v["ip"])
                record.get("value", []).sort(key=lambda v: v["ip"])
                for f in dns_update:
                    if record.get(f) != dns_update[f]:
                        needs_update = True
                        break
                if needs_update:
                    if len(servers) == 0:
                        yield f"Removing DNS record '{self.dns_name}'"
                        self.api.delete(
                            "domains", self.sld, "dns-records", record.get("id")
                        )
                    else:
                        yield f"Updating DNS record '{self.dns_name}' with {len(servers)} IPs"
                        self.api.put(
                            "domains",
                            self.sld,
                            "dns-records",
                            record.get("id"),
                            json=dns_update,
                        )
                else:
                    yield f"DNS Record '{self.dns_name}' is up to date"
                break
        else:
            if len(servers) > 0:
                yield f"Creating DNS record '{self.dns_name}' with {len(servers)} IPs"
                self.api.post("domains", self.sld, "dns-records", json=dns_update)
            else:
                yield f"DNS record for '{self.dns_name}' already removed"

    def provision(self) -> Generator:
        data = self.api.get("domains", self.sld)

        if not data["status"] == "active":
            for i in range(5):
                yield f"Checking ns records attempt [{i+1}/10]"
                resp = self.api.get("/domains", self.sld, "ns-keys/check")
                if resp["ns_status"] is True:
                    break
                if i == 0:
                    console.log(
                        "Domain is not active, please add the records as the [bold]authorative nameservers[/bold] for the domain"
                    )
                    for ns in data["ns_keys"]:
                        console.log(ns, style="orange1")
                time.sleep(10)
            else:
                raise ProvisioningError("Timed out waiting for ns records to update")

        yield "Fetching SSL status"
        ssl = self.api.get("/domains", self.sld, "ssl")

        if not ssl["certificates"]:
            # SSL certificate creation doesn't work when DNS records don't exist
            # So we make sure something exists before asking for a certificate
            yield "Checking DNS record exists"
            records = self.api.get("/domains", self.sld, "dns-records")
            for r in records:
                if r["type"] == "a" and r["name"] == self.dns_name:
                    break
            else:
                yield "Creating dummy DNS record"
                self.api.post(
                    "domains",
                    self.sld,
                    "dns-records",
                    json={
                        "type": "a",
                        "name": self.dns_name,
                        "value": [
                            {"ip": "1.2.3.4", "port": None, "weight": 1, "country": ""}
                        ],
                        "ttl": 120,
                        "cloud": True,
                        "upstream_https": "default",
                        "ip_filter_mode": {
                            "count": "multi",
                            "order": "rr",
                            "geo_filter": "none",
                        },
                    },
                )

            yield "Requesting certificate creation"
            self.api.patch(
                "/domains",
                self.sld,
                "ssl",
                json={
                    "certificate": "managed",
                },
            )

            yield "Waiting for certificates to be created"
            for i in range(1801):
                if i % 60 == 0:
                    ssl = self.api.get("/domains", self.sld, "ssl")
                    if ssl["certificates"]:
                        return True
                yield f"Waiting for certificates to be created [{i}/1800s]"
                time.sleep(1)

            raise ProvisioningError("Certificate creation timed out")

        yield "Activating SSL"
        if not ssl["ssl_status"]:
            self.api.patch(
                "/domains",
                self.sld,
                "ssl",
                json={
                    "ssl_status": True,
                },
            )

    @property
    def api(self):
        return self.provider.api


@dataclass
class ArvanProvider(CDNProvider):
    cloud = "ArvanCloud"
    domain_cls = ArvanDomain

    api_key: str = PromptField()

    _api = None

    def create_domain(self, domain_name: str) -> Generator:
        domain = ArvanDomain(name=domain_name, provider=self)

        yield "Checking domain status"

        api_domain = self.api.get("domains", self.sld)
        if api_domain is None:
            yield f"Registering new domain {domain.sld} with Arvan"
            api_domain = self.api.post(
                "/domains/dns-service",
                json={"domain": domain.sld, "domain_type": "full", "plan_level": 1},
            )
        return domain

    @property
    def api(self):
        if self._api is None:
            self._api = ArvanAPI(self.api_key)
        return self._api



import dataclasses
import random
import string
from dataclasses import field

import paramiko
from cryptography.hazmat.primitives import serialization


def ssh_key_fingerprint(key_path):
    k = paramiko.RSAKey.from_private_key_file(key_path)
    return ":".join(["%02x" % a for a in k.get_fingerprint()])


def ssh_key_public_key(key_path):
    k = paramiko.RSAKey.from_private_key_file(key_path)
    return (
        k.key.public_key()
        .public_bytes(
            serialization.Encoding.OpenSSH, serialization.PublicFormat.OpenSSH
        )
        .decode("utf-8")
    )


def random_string_lower(c=12) -> str:
    return "".join(random.choices(string.ascii_lowercase, k=c))


def random_string(c=12) -> str:
    return "".join(
        random.choices(
            string.ascii_lowercase + string.ascii_uppercase + string.digits, k=c
        )
    )


def dataclass(*args, **kwargs):
    """
    A wrapper around @dataclass. It removes all class fields starting with
    an underscore (_) from the fields list.
    """
    from dataclasses import dataclass as original_dataclass

    d = original_dataclass(*args, **kwargs)
    for k in [x for x in d.__dataclass_fields__ if x.startswith("_")]:
        d.__dataclass_fields__.pop(k)
    return d


def PromptField(type="input", required=True, **kwargs):
    return dataclasses.field(
        metadata={"prompt": {"type": type, "required": required, **kwargs}}
    )


def prompt_factory(cls):
    prompts = []
    for f in dataclasses.fields(cls):
        p = f.metadata.get("prompt")
        if not p:
            continue
        if not p.get("message"):
            p["message"] = f"Enter {f.name}"
        if not p.get("name"):
            p["name"] = f.name

        if "required" in p and p.pop("required"):
            v = p.get("validate")
            if v:
                p["validate"] = lambda r: r and v(r)
            else:
                p["validate"] = lambda r: r

        if "default_factory" in p:
            p["default"] = p.pop("default_factory")()

        prompts.append(p)
    return prompts



import os

import yaml
from vipen.core.task import with_prefix
from vipen.core.utils import dataclass
from vipen.providers import Server
from vipen.providers.base import Domain
from vipen.templates.base import script_install_docker, script_install_docker_compose


@dataclass
class Monitoring:
    server: Server
    domain: Domain

    monitoring_dir = "/root/monitoring"

    def provision_server(self, config):
        server, domain = self.server, self.domain

        if not server.is_connected:
            yield from with_prefix("Connecting to server", server.connect())

        yield from with_prefix("Installing Docker", server.run(script_install_docker))

        yield from with_prefix(
            "Installing Docker Compose", server.run(script_install_docker_compose)
        )

        yield from with_prefix(
            f"Fetching certificates for {domain.name}",
            server.run(
                f"""
                mkdir -p {self.monitoring_dir}
                cd {self.monitoring_dir}

                docker run --rm  -it  \
                    -v "{self.monitoring_dir}/acme":/acme.sh \
                    neilpang/acme.sh --register-account -m me@"{domain.sld}"

                docker run --rm  -it  \
                    --net=host \
                    -v "{self.monitoring_dir}/acme":/acme.sh \
                    neilpang/acme.sh  --issue -d {domain.name} --standalone
                res=$?
                if [ $res -eq 2 ]; then
                    echo "Certificates already exist"
                elif [ $res -ne 0 ]; then
                    exit $res
                fi

                mkdir {self.monitoring_dir}/cert
                cp {os.path.join(self.monitoring_dir, "acme", domain.name, "fullchain.cer")} {self.monitoring_dir}/cert/tls.crt
                cp {os.path.join(self.monitoring_dir, "acme", domain.name, f"{domain.name}.key")} {self.monitoring_dir}/cert/tls.key
                """
            ),
        )

        yield from self.refresh(config)

    def refresh(self, config):
        server, domain = self.server, self.domain

        prom_targets = []
        for cluster in config.clusters:
            for s in cluster.servers:
                prom_targets.append(
                    {
                        "targets": [s.ip],
                        "labels": {
                            "ip": s.ip,
                            "server": s.name,
                            "provider": s.provider.name,
                            "cluster": cluster.name,
                        },
                    }
                )

        from .dashboard_overview import dashboard as overview_dashboard
        from .dashboard_traffic import dashboard as traffic_dashboard

        yield from with_prefix(
            "Copying dashboard configurations",
            server.run(
                f"""
                    cd {self.monitoring_dir}

                    mkdir -p grafana/datasources grafana/dashboards
                    echo '{grafana_datasource}' > grafana/datasources/datasource.yaml
                    echo '{grafana_dashboard}' > grafana/dashboards/dashboard.yaml
                    echo '{overview_dashboard}' > grafana/dashboards/overview.json
                    echo '{traffic_dashboard}' > grafana/dashboards/traffic.json
                """,
            ),
        )

        yield from with_prefix(
            "Running Services",
            server.run(
                f"""
                    cd {self.monitoring_dir}
                    echo '{docker_compose}' > docker-compose.yml
                    echo '{prom_conf(prom_targets)}' > prometheus.yaml
                    echo '{nginx_config(domain=domain.name)}' > nginx.conf
                    docker-compose up -d
                """,
            ),
        )


docker_compose = """
version: "3"
services:
  prometheus:
    image: prom/prometheus:v2.40.1
    restart: always
    ports:
      - 9090:9090
    volumes:
      - ./prometheus.yaml:/etc/prometheus/prometheus.yml
      - prometheus:/prometheus
  grafana:
    image: grafana/grafana:8.2.6
    ports:
      - 3000:3000
    restart: always
    volumes:
      - grafana:/var/lib/grafana
      - ./grafana/datasources:/etc/grafana/provisioning/datasources
      - ./grafana/dashboards:/etc/grafana/provisioning/dashboards
  nginx:
    image: nginx
    restart: always
    ports:
      - 80:80
      - 443:443
    volumes:
      - ./cert:/root/cert
      - ./nginx.conf:/etc/nginx/conf.d/default.conf
volumes:
  grafana:
  prometheus:
""".strip()


prom_conf = lambda targets: yaml.safe_dump(
    {
        "global": {"scrape_interval": "15s", "evaluation_interval": "15s"},
        "scrape_configs": [
            {
                "job_name": "vipen-server",
                "relabel_configs": [
                    {
                        "source_labels": ["server"],
                        "target_label": "instance",
                    },
                ],
                "static_configs": targets,
            }
        ],
    }
)

nginx_config = (
    lambda domain: """
server {
    listen 443 ssl;
    server_name %s;
    ssl_certificate       /root/cert/tls.crt;
    ssl_certificate_key   /root/cert/tls.key;
    ssl_protocols         TLSv1 TLSv1.1 TLSv1.2;
    ssl_ciphers           HIGH:!aNULL:!MD5;

    location / {
        proxy_redirect off;
        proxy_pass http://grafana:3000;
        proxy_http_version 1.1;
    }
}

server {
    listen 80 default_server;
    server_name %s;
    return 301 https://$host$request_uri;
}
""".strip()
    % (domain, domain)
)

grafana_datasource = """
datasources:
  - name: Prometheus
    type: prometheus
    access: proxy
    url: http://prometheus:9090
    isDefault: true
    editable: true
""".strip()

grafana_dashboard = """
apiVersion: 1
providers:
  - name: 'Default'
    folder: 'vipen'
    options:
      path: /etc/grafana/provisioning/dashboards
""".strip()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['vipen',
 'vipen.cli',
 'vipen.core',
 'vipen.monitoring',
 'vipen.providers',
 'vipen.templates']

package_data = \
{'': ['*']}

install_requires = \
['deepdiff>=6.2.1,<7.0.0',
 'inquirerpy>=0.3.4,<0.4.0',
 'paramiko>=2.12.0,<3.0.0',
 'python-digitalocean>=1.17.0,<2.0.0',
 'pyyaml>=6.0,<7.0',
 'rich>=12.6.0,<13.0.0',
 'tld>=0.12.6,<0.13.0',
 'tqdm>=4.60.0,<5.0.0']

entry_points = \
{'console_scripts': ['vipen = vipen:main']}

setup_kwargs = {
    'name': 'vipen',
    'version': '0.0.1',
    'description': '',
    'long_description': 'None',
    'author': 'Hadi',
    'author_email': 'hadi.zolfaghaari@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import random
from email.generator import Generator

import click
from click import ClickException
from InquirerPy import prompt
from InquirerPy.base.control import Choice
from rich import box
from rich.table import Table
from vipen.cli.monitoring import monitoring
from vipen.core.logging import console
from vipen.core.task import TaskPool, run_task
from vipen.core.utils import random_string_lower

from .main import cli
from .util import choose_server_provider, current_cluster, get_config, new_progress


@cli.group()
@click.pass_context
def scale(ctx):
    # choose_server_provider(ctx, provider)
    pass


@scale.command(name="up")
@click.option("--provider", "-p", "provider_name", default=None)
@click.argument("count", nargs=1, default=1, type=int)
@click.pass_context
def scale_up(ctx, provider_name, count):
    config = get_config(ctx)
    cluster = current_cluster(ctx)
    provider = choose_server_provider(ctx, provider_name)

    if count > 10:
        raise ClickException("Scaling up limited to at most 10 servers at a time")

    def new_server(name) -> Generator:
        server = yield from provider.create_server(cluster, name)
        yield from cluster.template.provision_server(server)
        return server

    with new_progress() as p:
        pool = TaskPool(progress=p)
        with pool:
            for _ in range(count):
                name = f"{cluster.name}-{random_string_lower(4)}"
                pool.submit(name, new_server(name))

        cluster.servers.extend([s for s in pool.results if s is not None])

    with new_progress() as p:
        with TaskPool(progress=p) as pool:
            for domain in cluster.domains:
                pool.submit(
                    f"{domain.name}",
                    domain.update_dns_with_servers(cluster.servers),
                )

            if config.monitoring is not None:
                config.monitoring.refresh()


@scale.command(name="down")
@click.option("--provider", "-p", "provider_name", default=None)
@click.argument("count", nargs=1, default=1, type=int)
@click.pass_context
def scale_down(ctx, provider_name, count):
    config = get_config(ctx)
    cluster = current_cluster(ctx)

    if provider_name is not None:
        servers = [s for s in cluster.servers if s.provider.name == provider_name]
    else:
        servers = cluster.servers

    if count > len(servers):
        count = len(servers)

    servers = random.sample(servers, count)
    if provider_name is not None:
        console.log(f"Removing {len(servers)} servers from provider '{provider_name}'")
    else:
        console.log(f"Removing {len(servers)} servers")
    for i in range(count):
        cluster.servers.remove(servers[i])

    with new_progress() as p:
        with TaskPool(progress=p) as pool:
            for domain in cluster.domains:
                pool.submit(
                    domain.name,
                    domain.update_dns_with_servers(cluster.servers),
                )

    with new_progress() as p:
        with TaskPool(progress=p) as pool:
            for server in servers:
                pool.submit(server.pname, server.destroy())

            if config.monitoring is not None:
                config.monitoring.refresh()


@cli.group()
@click.pass_context
def server(ctx):
    pass


@server.command(name="rm")
@click.argument(
    "server_names",
    nargs=-1,
)
@click.pass_context
def remove_server(ctx, server_names):
    config = get_config(ctx)
    cluster = current_cluster(ctx)

    servers = []
    if len(server_names) > 0:
        for s in server_names:
            server = cluster.get_server(name=s)
            if server is None:
                console.error(f"Server {s} not found")
            servers.append(server)
    else:
        answer = prompt(
            {
                "message": "Which servers should I destroy? (You can select multiple server by pressing Space) ",
                "type": "list",
                "multiselect": True,
                "choices": [
                    Choice(
                        name=f"{cluster.servers[s].provider.name}/{cluster.servers[s].name}",
                        value=s,
                    )
                    for s in range(len(cluster.servers))
                ],
            },
        )
        servers = [cluster.servers[s] for s in answer[0]]

    for server in servers:
        cluster.servers.remove(server)

    with new_progress() as p:
        with TaskPool(progress=p) as pool:
            for domain in cluster.domains:
                pool.submit(
                    domain.name,
                    domain.update_dns_with_servers(cluster.servers),
                )

    with new_progress() as p:
        with TaskPool(progress=p) as pool:
            for server in servers:
                pool.submit(server.pname, server.destroy())

            if config.monitoring is not None:
                config.monitoring.refresh()


@server.command(name="ls")
@click.option("--all-clusters", "-a", type=bool, default=False, is_flag=True)
@click.pass_context
def list_servers(ctx, all_clusters):
    config = get_config(ctx)

    table = Table(box=box.SIMPLE_HEAVY, show_header=False, show_footer=False)

    table.add_column("Cloud", justify="right", style="cyan", no_wrap=True)
    table.add_column("Provider", justify="right", style="cyan", no_wrap=True)
    table.add_column("Name", style="magenta")
    table.add_column("IP", style="magenta")

    if all_clusters:
        clusters = config.clusters
    else:
        clusters = [current_cluster(ctx)]

    for cluster in clusters:
        for s in cluster.servers:
            table.add_row(s.provider.cloud, s.provider.name, s.name, s.ip)

    console.print(table)



import base64
import json
import urllib.parse
from dataclasses import dataclass
from uuid import uuid4

from vipen.core.config import Cluster
from vipen.core.task import with_prefix
from vipen.core.utils import PromptField, random_string
from vipen.providers import Domain, Server

from .base import (
    Template,
    script_install_docker,
    script_install_docker_compose,
    tls_crt,
    tls_key,
)


@dataclass
class ShadowsocksV2RayPlugin(Template):
    type = "ss_plugin_v2ray"
    description = "Shadowsocks with V2Ray Plugin"

    password: str = PromptField(default_factory=lambda: random_string(16))
    method: str = PromptField(default="chacha20-ietf-poly1305")

    vipen_dir = "/root/vipen"
    v2ray_path = "/vipen"

    def provision_server(self, server: Server):
        if not server.is_connected:
            yield from with_prefix("Connecting to server", server.connect())

        yield from with_prefix("Installing Docker", server.run(script_install_docker))

        yield from with_prefix(
            "Installing Docker Compose", server.run(script_install_docker_compose)
        )

        yield from with_prefix(
            "Running Services",
            server.run(
                f"""
                    mkdir -p {self.vipen_dir}
                    cd {self.vipen_dir}

                    mkdir cert
                    echo '{tls_crt}' > cert/tls.crt
                    echo '{tls_key}' > cert/tls.key

                    echo '{docker_compose(password=self.password, method=self.method, path=self.v2ray_path)}' > docker-compose.yml
                    echo '{nginx_config(path=self.v2ray_path)}' > nginx.conf
                    docker-compose up -d
                """,
            ),
        )

    def generate_key(self, cluster: Cluster, domain: Domain):
        secret = base64.b64encode(f"{self.method}:{self.password}").decode("utf-8")
        params = {
            "plugin": "v2ray-plugin",
            "path": self.v2ray_path,
            "loglevel": "none",
            "host": domain.name,
        }
        name = f"{cluster.name}-{domain.sld}"
        return f"ss://{secret}@{domain.name}:443/?{urllib.parse.urlencode(params)}&tls#{name}"


docker_compose = (
    lambda path, password, method: """
version: "3"
services:
  shadowsocks:
    image: acrisliu/shadowsocks-libev
    environment:
        - ARGS="--plugin v2ray-plugin --plugin-opts server;path=%s -u"
        - PASSWORD="%s"
        - METHOD="%s"
    restart: always
  wsproxy:
    image: ghcr.io/ghitf/wsproxy:main
    command: ["--backend", "ws://shadowsocks:8388", "--metrics", "0.0.0.0:9900"]
    restart: always
  nginx:
    image: nginx
    restart: always
    ports:
      - 80:80
      - 443:443
    volumes:
      - ./cert:/root/cert
      - ./nginx.conf:/etc/nginx/conf.d/default.conf
""".strip()
    % (path, password, method)
)

nginx_config = (
    lambda path: """
server {
    listen 443 ssl;
    ssl_certificate       /root/cert/tls.crt;
    ssl_certificate_key   /root/cert/tls.key;
    ssl_protocols         TLSv1 TLSv1.1 TLSv1.2;
    ssl_ciphers           HIGH:!aNULL:!MD5;

    location %s {
      if ($http_upgrade != "websocket") { # Return 404 error when WebSocket upgrading negotiate failed
          return 404;
      }
      proxy_redirect off;
      proxy_pass http://wsproxy;
      proxy_http_version 1.1;
      proxy_set_header Upgrade $http_upgrade;
      proxy_set_header Connection "upgrade";
      proxy_set_header Host $host;
      # Show real IP in v2ray access.log
      proxy_set_header X-Real-IP $remote_addr;
      proxy_set_header X-Forwarded-For $proxy_add_x_forwarded_for;
    }

    location /metrics {
        proxy_redirect off;
        proxy_pass http://wsproxy:9900;
        proxy_http_version 1.1;
    }

    location / {
        try_files $uri $uri/ =404;
    }
}

server {
    listen 80;

    location %s {
      if ($http_upgrade != "websocket") { # Return 404 error when WebSocket upgrading negotiate failed
          return 404;
      }
      proxy_redirect off;
      proxy_pass http://wsproxy;
      proxy_http_version 1.1;
      proxy_set_header Upgrade $http_upgrade;
      proxy_set_header Connection "upgrade";
      proxy_set_header Host $host;
      # Show real IP in v2ray access.log
      proxy_set_header X-Real-IP $remote_addr;
      proxy_set_header X-Forwarded-For $proxy_add_x_forwarded_for;
    }

    location /metrics {
        proxy_redirect off;
        proxy_pass http://wsproxy:9900;
        proxy_http_version 1.1;
    }

    location / {
        try_files $uri $uri/ =404;
    }
}
""".strip()
    % (path, path)
)



import click
from rich import box
from rich.table import Table
from vipen.core.config import Cluster
from vipen.core.logging import console
from vipen.core.task import run_task

from .main import cli
from .util import choose_cdn_provider, current_cluster, new_progress


@cli.group()
def domain():
    pass


@domain.command(name="add")
@click.option("--provider", "-p", default=None)
@click.argument(
    "domain_name",
    nargs=1,
)
@click.pass_context
def add_domain(ctx, provider, domain_name):
    cluster = current_cluster(ctx)
    cdn_provider = choose_cdn_provider(ctx, provider)

    def provision(domain_name):
        domain = yield from cdn_provider.create_domain(domain_name)
        yield from domain.provision()
        yield from domain.update_dns_with_servers(cluster.servers)
        return domain

    with new_progress() as p:
        domain = run_task(domain_name, provision(domain_name), progress=p)

    if domain:
        cluster.domains.append(domain)


@domain.command(name="rm")
@click.argument(
    "domain_name",
    nargs=1,
)
@click.pass_context
def remove_domain(ctx, domain_name):
    cluster: Cluster = current_cluster(ctx)

    domain = cluster.get_domain(domain_name)
    if domain is None:
        console.error(f"Domain {domain_name} not found")
        return

    console.log(f"Removing domain '{domain_name}'.")
    console.log(
        "This only removes it from the cluster config, if you want it removed from the provider you must do that manually."
    )
    cluster.domains.remove(domain)


@domain.command(name="ls")
@click.pass_context
def list_domains(ctx):
    cluster: Cluster = current_cluster(ctx)

    table = Table(box=box.SIMPLE_HEAVY, show_header=False, show_footer=False)

    table.add_column("Cloud", justify="right", style="cyan", no_wrap=True)
    table.add_column("Provider", justify="right", style="cyan", no_wrap=True)
    table.add_column("Name", style="magenta")

    for domain in cluster.domains:
        table.add_row(domain.provider.cloud, domain.provider.name, domain.name)

    console.print(table)



import os

import click
from click import ClickException
from vipen.core.config import DEFAULT_CONFIG_DIR, ConfigError, load_config

from .util import get_config


@click.group()
@click.option(
    "--config", "config_path", default=os.path.join(DEFAULT_CONFIG_DIR, "config.yaml")
)
@click.option("--cluster", "-c", default=None)
@click.pass_context
def cli(ctx, config_path, cluster):
    from vipen.providers.arvan import ArvanProvider
    from vipen.providers.digitalocean import DigitalOceanProvider
    from vipen.templates.ss_v2ray_plugin import ShadowsocksV2RayPlugin
    from vipen.templates.v2ray_ws_tls_noaead import V2RayWSTlsNoAEAD

    if not os.path.exists(DEFAULT_CONFIG_DIR):
        os.makedirs(DEFAULT_CONFIG_DIR)

    ctx.obj = {}

    try:
        ctx.obj["config"] = load_config(
            config_path,
            use_cluster=cluster,
            provider_cls=[DigitalOceanProvider, ArvanProvider],
            template_cls=[V2RayWSTlsNoAEAD, ShadowsocksV2RayPlugin],
        )
    except ConfigError as e:
        raise ClickException(str(e))


@cli.result_callback()
@click.pass_context
def save_config(ctx, *args, **kwargs):
    get_config(ctx).save(kwargs.get("config_path"))



from rich.console import Console

console = Console()

console.warn = lambda *args: console.log(*args)
console.error = lambda *args: console.log(*args)
console.debug = lambda *args: None  # console.log(*args)
